import mock
import pexpect
import pytest

from ocflib.account.utils import extract_username_from_principal
from ocflib.account.utils import get_vhost_db
from ocflib.account.utils import get_vhosts
from ocflib.account.utils import has_vhost
from ocflib.account.utils import home_dir
from ocflib.account.utils import password_matches
from ocflib.account.utils import web_dir


class TestPasswordMatches:

    @pytest.mark.parametrize('user', [
        '',
        '; rm -rf /',
        ';',
        '&&',
        ';exit',
        '\\',
        'thisistoolong',
    ])
    def test_fails_with_bad_username(self, user):
        with pytest.raises(ValueError):
            password_matches(user, 'hunter2')

    @pytest.mark.parametrize('password', [
        'asdf\nasdf',
        '\n\n\n\n\n\n',
        'a\0asdf',
    ])
    def test_fails_with_bad_password(self, password):
        with pytest.raises(ValueError):
            password_matches('ckuehl', password)

    @mock.patch('ocflib.account.validators.user_exists', return_value=False)
    def test_fails_if_user_does_not_exist(self, __):
        with pytest.raises(ValueError):
            password_matches('ckuehl', 'hunter2')

    @mock.patch('ocflib.account.validators.user_exists', return_value=True)
    @mock.patch('pexpect.spawn')
    def test_calls_pexpect_correctly(self, spawn, __):
        password_matches('ckuehl', 'hunter2')
        spawn.assert_called_with(
            'kinit --no-forwardable -l0 ckuehl@OCF.BERKELEY.EDU',
            timeout=10,
        )

        child = spawn.return_value

        assert child.expect.mock_calls == [
            mock.call.first('ckuehl@OCF.BERKELEY.EDU\'s Password:'),
            mock.call.second(pexpect.EOF),
        ]
        child.sendline.assert_called_with('hunter2')
        assert child.close.called

    @mock.patch('ocflib.account.validators.user_exists', return_value=True)
    @mock.patch('pexpect.spawn')
    @pytest.mark.parametrize('exitstatus,success', [
        (0, True),

        # this is excessive, but we want to be really sure we don't
        # accidentally accept a bad password
        (1, False),
        (22, False),
        (-1, False),
        (None, False),
    ])
    def test_returns_correctly(self, spawn, __, exitstatus, success):
        spawn.return_value.exitstatus = exitstatus
        assert password_matches('ckuehl', 'hunter2') == success


class TestExtractUsernameFromPrincipal:

    @pytest.mark.parametrize('principal,username', [
        ('ckuehl@OCF.BERKELEY.EDU', 'ckuehl'),
        ('ckuehl/admin@OCF.BERKELEY.EDU', 'ckuehl'),
        ('ckuehl/root@OCF.BERKELEY.EDU', 'ckuehl'),
    ])
    def test_success(self, principal, username):
        assert extract_username_from_principal(principal) == username

    @pytest.mark.parametrize('principal', [
        '',
        'ckuehl',
        'ckuehl@ocf.berkeley.edu',
        '@OCF.BERKELEY.EDU',
        '@',
        'thisiswaytoolongn@OCF.BERKELEY.EDU',
        'ckuehl+root@OCF.BERKELEY.EDU',
    ])
    def test_fails_on_bad_input(self, principal):
        with pytest.raises(ValueError):
            extract_username_from_principal(principal)


VHOSTS_EXAMPLE = """
# [added 2015.05.05 ckuehl]
asucarch archive.asuc.org www.archive.asuc.org,modern.asuc.org,www.modern.asuc.org -

# [added 2015.04.16 ckuehl]
staff! contrib - /contrib
ocfwiki docs.ocf.berkeley.edu - - [ssl,hsts]
"""  # noqa

VHOSTS_EXAMPLE_PARSED = {
    'archive.asuc.org': {
        'aliases': [
            'www.archive.asuc.org',
            'modern.asuc.org',
            'www.modern.asuc.org',
        ],
        'docroot': '/',
        'flags': [],
        'redirect': None,
        'username': 'asucarch',
    },
    'contrib.berkeley.edu': {
        'aliases': [],
        'docroot': '/contrib',
        'flags': [],
        'redirect': '/ https://www.ocf.berkeley.edu/~staff/',
        'username': 'staff',
    },
    'docs.ocf.berkeley.edu': {
        'aliases': [],
        'docroot': '/',
        'flags': ['ssl', 'hsts'],
        'redirect': None,
        'username': 'ocfwiki',
    },
}


@pytest.yield_fixture
def mock_get_vhosts_db():
    with mock.patch(
        'ocflib.account.utils.get_vhost_db',
        return_value=VHOSTS_EXAMPLE.splitlines()
    ):
        yield


class TestVirtualHosts:

    def test_reads_file_if_exists(self):
        with mock.patch('builtins.open', mock.mock_open()) as mock_open:
            lines = ['hello', 'world']
            mock_open.return_value.__iter__.return_value = lines
            assert get_vhost_db() == lines

    @mock.patch('builtins.open')
    @mock.patch('requests.get')
    def test_reads_web_if_no_file(self, get, mock_open):
        def raise_error(__):
            raise IOError()

        mock_open.side_effect = raise_error
        get.return_value.text = 'hello\nworld'

        assert get_vhost_db() == ['hello', 'world']

    def test_proper_parse(self, mock_get_vhosts_db):
        assert get_vhosts() == VHOSTS_EXAMPLE_PARSED

    @pytest.mark.parametrize('user,should_have_vhost', [
        ('staff', True),
        ('ocfwiki', True),
        ('ckuehl', False),
        ('', False),
    ])
    def test_has_vhost(self, user, should_have_vhost, mock_get_vhosts_db):
        assert has_vhost(user) == should_have_vhost


class TestUserPaths:

    @pytest.mark.parametrize('user,expected', [
        ('ckuehl', '/home/c/ck/ckuehl'),
        ('ggroup', '/home/g/gg/ggroup'),
    ])
    def test_home_dir(self, user, expected):
        assert home_dir(user) == expected

    @pytest.mark.xfail
    @pytest.mark.parametrize('user', [
        '',
        'c',
        'ck',
    ])
    def test_home_dir_errors_bad_user(self, user):
        with pytest.raises(ValueError):
            home_dir(user)

    @pytest.mark.parametrize('user,expected', [
        ('ckuehl', '/services/http/users/c/ckuehl'),
        ('ggroup', '/services/http/users/g/ggroup'),
    ])
    def test_web_dir(self, user, expected):
        assert web_dir(user) == expected

    @pytest.mark.xfail
    @pytest.mark.parametrize('user', [
        '',
        'c',
        'ck',
    ])
    def test_web_dir_errors_bad_user(self, user):
        with pytest.raises(ValueError):
            web_dir(user)



import pytest

from ocflib.misc.validators import email_host_exists
from ocflib.misc.validators import host_exists
from ocflib.misc.validators import valid_email


REAL_HOSTS = ['ocf.berkeley.edu', 'google.com']
FAKE_HOSTS = [
    'hello.waffle',
    'i am not a real host',
    'lashdfkjashgklhsadfsad.com',
]


@pytest.mark.parametrize(
    'host,exists',
    [(host, True) for host in REAL_HOSTS] +
    [(host, False) for host in FAKE_HOSTS]
)
def test_host_exists(host, exists):
    assert host_exists(host) == exists


@pytest.mark.parametrize(
    'email,exists',
    [('ckuehl@' + host, True) for host in REAL_HOSTS] +
    [('ckuehl@' + host, False) for host in FAKE_HOSTS] +
    [('no host here!', False)],
)
def test_email_host_exists(email, exists):
    assert email_host_exists(email) == exists


@pytest.mark.parametrize('email,valid', [
    ('ckuehl@ocf.berkeley.edu', True),
    ('hello-world_i+am.email@google.com', True),
    ('hello world@ocf.berkeley.edu', False),
    ('hello world@ocf', False),
    ('derp@langasdgkjsadhglkjbjgsdfgsd.com', False),
    ('derp@www.ocf.berkeley.edu', False),  # no MX records
    ('@ocf', False),
    ('hello@', False),
    ('@', False),
    ('', False),
])
def test_valid_email(email, valid):
    assert valid_email(email) == valid



import pytest

from ocflib.infra.hosts import hosts_by_filter


class TestHostsByFilter:

    def _hostnames(self, results):
        return [entry['cn'][0] for entry in results]

    @pytest.mark.parametrize('filter_str,expected', [
        ('(cn=death)', ['death']),
        ('(cn=doesnotexist)', []),
        ('(herp=derp)', []),
    ])
    def test_users_by_filter(self, filter_str, expected):
        results = self._hostnames(hosts_by_filter(filter_str))
        assert set(results) == set(expected)

    @pytest.mark.parametrize('filter_str', ['', 'cn=death', '42', 'asdf'])
    def test_invalid_filters(self, filter_str):
        with pytest.raises(Exception):
            hosts_by_filter(filter_str)

    def test_puppet_class(self):
        # This will break if death is ever renamed, but it's a useful test.
        #
        # We choose to test death because it is in lots of university DNS
        # records, so it is probably one of the more unlikely hosts to be
        # renamed.
        assert ('death' in
                self._hostnames(hosts_by_filter('(puppetClass=ocf_www)')))



from contextlib import contextmanager

import mock
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from ocflib.account.submission import Base
from ocflib.account.submission import get_tasks
from ocflib.account.submission import StoredNewAccountRequest
from ocflib.account.submission import user_has_request_pending
from ocflib.account.submission import username_pending
from tests.account.creation_test import fake_credentials  # noqa
from tests.account.creation_test import fake_new_account_request  # noqa
from tests.account.creation_test import mock_rsa_key  # noqa


@pytest.fixture
def session(fake_credentials):
    engine = create_engine(fake_credentials.mysql_uri)
    Base.metadata.create_all(engine)
    return sessionmaker(bind=engine)()


class TestUsernamePending:

    def test_pending(self, session, fake_new_account_request):
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request))
        session.commit()
        assert username_pending(session, fake_new_account_request)

    def test_not_pending(self, session, fake_new_account_request):
        assert not username_pending(session, fake_new_account_request)


class TestUserHasRequestPending:

    @pytest.mark.parametrize('attrs', [
        {'is_group': True, 'calnet_uid': None, 'callink_oid': 14},
        {'is_group': False, 'calnet_uid': 14, 'callink_oid': None},
    ])
    def test_has_pending(self, session, fake_new_account_request, attrs):
        fake_new_account_request = fake_new_account_request._replace(**attrs)
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request))
        session.commit()
        assert user_has_request_pending(session, fake_new_account_request)

    def test_not_has_pending(self, session, fake_new_account_request):
        assert not user_has_request_pending(session, fake_new_account_request)

    def test_not_has_pending_zero_group(self, session, fake_new_account_request):
        """callink_oid=0 can create infinite accounts."""
        fake_new_account_request = fake_new_account_request._replace(
            is_group=True,
            callink_oid=0,
            calnet_uid=None,
        )
        session.add(StoredNewAccountRequest.from_request(fake_new_account_request))
        session.commit()
        assert not user_has_request_pending(session, fake_new_account_request)


@pytest.fixture
def celery_app():
    sent_messages = []

    def mock_celery_task(f):
        def wrapper(*args, **kwargs):
            f(*args, **kwargs)

        def update_state(**kwargs):
            pass

        # wrap everything in Mock so we can track calls
        return mock.Mock(
            side_effect=wrapper,
            delay=mock.Mock(side_effect=wrapper),
            update_state=mock.Mock(side_effect=update_state),
        )

    @contextmanager
    def dispatcher():
        def send(**kwargs):
            sent_messages.append(kwargs)
        yield mock.Mock(send=send)

    return mock.Mock(
        task=mock_celery_task,
        _sent_messages=sent_messages,
        **{
            'events.default_dispatcher': dispatcher,
        }
    )


@pytest.yield_fixture
def tasks(session, celery_app, fake_credentials):
    with mock.patch('ocflib.account.submission.sessionmaker', return_value=lambda: session):
        yield get_tasks(celery_app, credentials=fake_credentials)


@pytest.fixture
def session_with_requests(session, fake_new_account_request):
    session.add(StoredNewAccountRequest.from_request(fake_new_account_request))
    session.add(StoredNewAccountRequest.from_request(fake_new_account_request._replace(user_name='other')))
    session.commit()
    assert len(session.query(StoredNewAccountRequest).all()) == 2
    return session


def test_approve_request(celery_app, fake_new_account_request, session_with_requests, tasks):
    tasks.approve_request(fake_new_account_request.user_name)
    assert len(session_with_requests.query(StoredNewAccountRequest).all()) == 1

    # we want to make sure we go through the same conversion process:
    # live -> stored -> live -> dict
    request = StoredNewAccountRequest.from_request(fake_new_account_request).to_request()
    tasks.create_account.delay.assert_called_once_with(request)
    assert celery_app._sent_messages == [
        {'type': 'ocflib.account_approved', 'request': request.to_dict()}
    ]


@mock.patch('ocflib.account.submission.send_rejected_mail')
def test_reject_request(send_rejected_mail, celery_app, fake_new_account_request, session_with_requests, tasks):
    tasks.reject_request(fake_new_account_request.user_name)
    assert len(session_with_requests.query(StoredNewAccountRequest).all()) == 1

    # we want to make sure we go through the same conversion process:
    # live -> stored -> live -> dict
    request = StoredNewAccountRequest.from_request(fake_new_account_request).to_request()
    assert celery_app._sent_messages == [
        {'type': 'ocflib.account_rejected', 'request': request.to_dict()}
    ]

    # TODO: assert real reason
    send_rejected_mail.assert_called_once_with(request, mock.ANY)



import os.path
from contextlib import contextmanager
from textwrap import dedent

import mock
import pytest

import ocflib.account.creation
import ocflib.constants as constants
from ocflib.account.creation import _get_first_available_uid
from ocflib.account.creation import create_account
from ocflib.account.creation import create_home_dir
from ocflib.account.creation import create_web_dir
from ocflib.account.creation import decrypt_password
from ocflib.account.creation import eligible_for_account
from ocflib.account.creation import encrypt_password
from ocflib.account.creation import NewAccountRequest
from ocflib.account.creation import send_created_mail
from ocflib.account.creation import send_rejected_mail
from ocflib.account.creation import validate_callink_oid
from ocflib.account.creation import validate_calnet_uid
from ocflib.account.creation import validate_email
from ocflib.account.creation import validate_password
from ocflib.account.creation import validate_request
from ocflib.account.creation import validate_username
from ocflib.account.creation import ValidationError
from ocflib.account.creation import ValidationWarning
from ocflib.account.submission import AccountCreationCredentials


WEAK_KEY = dedent(
    """\
    -----BEGIN RSA PRIVATE KEY-----
    MIICWwIBAAKBgQDGkGNFk/yy8HphSvvsmCpMF1vGbJeZXw2AmlLfTLcGJkZuvelu
    qJTGepGjeEeML6GrE03dI330mWtnC8jdhmwaELndqoPQ3Ks1eXF5usvDeYoRVir0
    ekqJtd2+7eBQ4xrRIA5YohoE31VGQ7ZaQ0GLMuWjldTe3bx+5OJqB0pE5QIDAQAB
    AoGAZtTX1GyzbagEeOZwWTLklMt0B+qtCAyl3XgOev4rus+PokJP5bMAeVl4mwPr
    aboxK3uv01pSHJ5ndNIxkCfRSzSuKvddfoaDrP97dbS0boqHyJmG38U2kxaMufBP
    rFP4a05TajdU9/rQSaGGmTkgDmRfJId5aDfJh6ToKMEYnQECQQDYb0Nj9mAWz5ja
    btXyegMespiK1UnyrZlRKsm0zcnEZ4NBE/lgMiJJXkfhdS5af9ORPmDjlQfwuHtZ
    N5mEKXNRAkEA6tzQPWCIL3gz0FYVX+l00JTFRIFA1yfvHiF4XjNZUr1TjXdGhto5
    DqV39XTk1CPtXNJto9AmNLf8zJD5xsqLVQJAToXnfD/p0rzUpwMpSgSsVxnSsCP7
    5TjIdCNC9P7oYgJwI0883YKy38195LVf8OOJfZuVCVyLefFkhxTd9I4ZUQJAO0ft
    D/DzrveqLGXuEz18DMHgYQA2+5fK1VIhbbWMUEQVeNmoZZVjXX0KoFwW/izrVsiO
    gBCj9B6UopXdVf392QJAcCUgxV6Ca6iWplAKHsaJ7sG+aQOaYI8m3d3MqJ5g34GB
    CqXzvT0v5ZrGj+K9dWDb+pYvGWhc2iU/e40yyj0G9w==
    -----END RSA PRIVATE KEY-----"""
)


@pytest.yield_fixture
def fake_new_account_request(mock_rsa_key):
    yield NewAccountRequest(
        user_name='someuser',
        real_name='Some User',
        is_group=False,
        calnet_uid=123456,
        callink_oid=None,
        email='some.user@ocf.berkeley.edu',
        encrypted_password=encrypt_password('hunter2000', mock_rsa_key),
        handle_warnings=NewAccountRequest.WARNINGS_WARN,
    )


@pytest.yield_fixture
def mock_rsa_key(tmpdir):
    test_key = tmpdir.join('test.key')
    test_key.write((WEAK_KEY + '\n').encode('ascii'))
    yield test_key.strpath
    test_key.remove()


class TestFirstAvailableUID:

    def test_first_uid(self):
        connection = mock.Mock(response=[
            {'attributes': {'uidNumber': [100]}},
            {'attributes': {'uidNumber': [300]}},
            {'attributes': {'uidNumber': [200]}},
        ])

        @contextmanager
        def ldap_ocf():
            yield connection

        with mock.patch('ocflib.account.creation.ldap_ocf', ldap_ocf):
            next_uid = _get_first_available_uid()

        connection.search.assert_called_with(
            constants.OCF_LDAP_PEOPLE,
            '(uidNumber=*)',
            attributes=['uidNumber'],
        )

        assert next_uid == 301


class TestCreateDirectories:

    @mock.patch('subprocess.check_call')
    def test_create_home_dir(self, check_call):
        create_home_dir('ckuehl')

        calls = [mock.call(
            ['sudo', 'install', '-d', '--mode=0700', '--group=ocf',
                '--owner=ckuehl', '/home/c/ck/ckuehl'],
        )]

        for name in ['bashrc', 'bash_profile', 'bash_logout']:
            path = os.path.join(os.path.dirname(
                ocflib.account.creation.__file__),
                'rc',
                name,
            )
            calls.append(mock.call(
                ['sudo', 'install', '--mode=0600', '--group=ocf',
                    '--owner=ckuehl', path, '/home/c/ck/ckuehl/.' + name],
            ))

        check_call.assert_has_calls(calls)

    @mock.patch('subprocess.check_call')
    def test_create_web_dir(self, check_call):
        create_web_dir('ckuehl')
        check_call.assert_called_with(
            ['sudo', 'install', '-d', '--mode=0000', '--group=ocf',
                '--owner=ckuehl', '/services/http/users/c/ckuehl'],
        )


class TestUsernameBasedOnRealName:

    @pytest.mark.parametrize('username,realname,success', [
        ['ckuehl', 'Christopher Kuehl', True],
        ['ckuehl', 'CHRISTOPHER B KUEHL', True],
        ['kuehl', 'CHRISTOPHER B KUEHL', True],
        ['cbk', 'CHRISTOPHER B KUEHL', True],

        ['rejectme', 'Christopher Kuehl', False],
        ['penguin', 'Christopher Kuehl', False],
        ['daradib', 'Christopher Kuehl', False],
    ])
    @mock.patch('ocflib.account.validators.validate_username')
    @mock.patch('ocflib.account.search.user_exists', return_value=False)
    def test_some_names(self, _, __, username, realname, success,):
        """Test some obviously good and bad usernames."""
        try:
            validate_username(username, realname)
        except ValidationWarning as ex:
            if success:
                pytest.fail(
                    'Received unexpected error: {error}'.format(error=ex),
                )

    @pytest.mark.parametrize('username', [
        'shitup',
        'hella',
        'ucbcop',
        'suxocf',
    ])
    @mock.patch('ocflib.account.search.user_exists', return_value=False)
    @mock.patch('ocflib.account.creation.similarity_heuristic', return_value=0)
    def test_warning_names(self, _, __, username):
        """Ensure that we raise warnings when bad/restricted words appear."""
        with pytest.raises(ValidationWarning):
            validate_username(username, username)

    @pytest.mark.parametrize('username', [
        'wordpress',
        'systemd',
        'ocf',
        'ocfrocks',
    ])
    @mock.patch('ocflib.account.search.user_exists', return_value=False)
    @mock.patch('ocflib.account.creation.similarity_heuristic', return_value=0)
    def test_error_names(self, _, __, username):
        """Ensure that we raise errors when appropriate."""
        with pytest.raises(ValidationError):
            validate_username(username, username)

    def test_error_user_exists(self):
        """Ensure that we raise an error if the username already exists."""
        with pytest.raises(ValidationError):
            validate_username('ckuehl', 'Chris Kuehl')

    @mock.patch('ocflib.account.validators.validate_username')
    @mock.patch('ocflib.account.search.user_exists', return_value=False)
    def test_long_names(self, _, __):
        """In the past, create has gotten "stuck" trying millions of
        combinations of real names because we try permutations of the words in
        the real name."""
        with pytest.raises(ValidationWarning):
            # 16! = 2.09227899e13, so if this works, it's definitely not
            # because we tried all possibilities
            validate_username(
                'nomatch',
                'I Have Sixteen Names A B C D E F G H I J K L',
            )


class TestAccountEligibility:

    @pytest.mark.parametrize('bad_uid', [
        1034192,     # good uid, but already has account
        9999999999,  # fake uid, not in university ldap
    ])
    def test_validate_calnet_uid_error(self, bad_uid):
        with pytest.raises(ValidationError):
            validate_calnet_uid(bad_uid)

    def test_validate_calnet_uid_success(self, mock_valid_calnet_uid):
        validate_calnet_uid(9999999999999)

    def test_validate_calnet_affiliations_failure(self, mock_invalid_calnet_uid):
        with pytest.raises(ValidationWarning):
            validate_calnet_uid(9999999999999)

    @pytest.mark.parametrize('affiliations,eligible', [
        (['AFFILIATE-TYPE-CONSULTANT'], True),
        (['AFFILIATE-TYPE-CONSULTANT', 'AFFILIATE-STATUS-EXPIRED'], False),

        (['EMPLOYEE-TYPE-ACADEMIC'], True),
        (['EMPLOYEE-TYPE-STAFF'], True),
        (['EMPLOYEE-TYPE-ACADEMIC', 'EMPLOYEE-STATUS-EXPIRED'], False),
        (['EMPLOYEE-STATUS-EXPIRED', 'AFFILIATE-TYPE-CONSULTANT'], True),
        (['EMPLOYEE-STATUS-EXPIRED', 'STUDENT-TYPE-REGISTERED'], True),

        (['STUDENT-TYPE-REGISTERED'], True),
        (['STUDENT-TYPE-REGISTERED', 'STUDENT-STATUS-EXPIRED'], False),
        (['STUDENT-STATUS-EXPIRED'], False),

        ([], False),
    ])
    def test_affiliations(self, affiliations, eligible):
        assert eligible_for_account(affiliations) == eligible


class TestSendMail:

    @mock.patch('ocflib.misc.mail.send_mail')
    def test_send_created_mail(self, send_mail, fake_new_account_request):
        send_created_mail(fake_new_account_request)
        send_mail.assert_called_once_with(
            fake_new_account_request.email,
            '[OCF] Your account has been created!',
            mock.ANY,
        )

    @mock.patch('ocflib.misc.mail.send_mail')
    def test_send_rejected_mail(self, send_mail, fake_new_account_request):
        send_rejected_mail(fake_new_account_request, 'some reason')
        send_mail.called_called_once_with(
            fake_new_account_request.email,
            '[OCF] Your account has been created!',
            mock.ANY,
        )


class TestPasswordEncryption:

    @pytest.mark.parametrize('password', [
        'hello world',
        'hunter2',
        'mock_send_mail_user.assert_called_once_with',
    ])
    def test_encrypt_decrypt_password(self, password, mock_rsa_key):
        assert decrypt_password(
            encrypt_password(password, mock_rsa_key),
            mock_rsa_key,
        ) == password


class TestValidateCallinkOid:

    @pytest.mark.parametrize('oid', [0, 123123123])
    def test_valid_oid(self, oid):
        validate_callink_oid(oid)

    @pytest.mark.parametrize('oid', [46130, 46187])
    def test_invalid_oid(self, oid):
        with pytest.raises(ValidationWarning):
            validate_callink_oid(oid)


class TestValidateEmail:

    @pytest.mark.parametrize('email', [
        'ckuehl@ocf.berkeley.edu',
        'somebody@gmail.com',
        'herp.derp-hello+something@berkeley.edu',
    ])
    def test_valid_email(self, email):
        validate_email(email)

    @pytest.mark.parametrize('email', [
        '',
        '@',
        'hello@hello',
        'some kinda email@gmail.com',
    ])
    def test_invalid_email(self, email):
        with pytest.raises(ValidationError):
            validate_email(email)


class TestValidatePassword:

    @pytest.mark.parametrize('password', [
        'correct horse battery staple',
        'pogjpaioshfoasdfnlka;sdfi;sagj',
        'p@ssw0rd',
    ])
    def test_valid_password(self, password):
        validate_password('ckuehl', password)

    @pytest.mark.parametrize('password', [
        '',
        'simple',
        'correct horse\nbattery staple',
        'correct horse battery staple é',
    ])
    def test_invalid_password(self, password):
        with pytest.raises(ValidationError):
            validate_password('ckuehl', password)


def mock_session(response):
    # TODO: this is crappy, use the one from submission_test instead
    return mock.Mock(**{
        'query.return_value': mock.Mock(
            scalar=lambda: response[0][0],
            first=lambda: response[0]
        ),
    })


@pytest.yield_fixture
def fake_credentials(mock_rsa_key):
    yield AccountCreationCredentials(
        encryption_key=mock_rsa_key,
        mysql_uri='sqlite://',  # heh
        kerberos_keytab='/nonexist',
        kerberos_principal='create/admin',
    )


@pytest.yield_fixture
def mock_valid_calnet_uid():
    with mock.patch(
        'ocflib.account.search.user_attrs_ucb',
        return_value={'berkeleyEduAffiliations': ['STUDENT-TYPE-REGISTERED']}
    ):
        yield


@pytest.yield_fixture
def mock_invalid_calnet_uid():
    with mock.patch(
        'ocflib.account.search.user_attrs_ucb',
        return_value={'berkeleyEduAffiliations': ['STUDENT-STATUS-EXPIRED']},
    ):
        yield


class TestValidateRequest:

    def test_valid_request(
        self,
        fake_new_account_request,
        fake_credentials,
        mock_valid_calnet_uid,
    ):
        assert validate_request(
            fake_new_account_request,
            fake_credentials,
            mock_session([[False]]),
        ) == ([], [])

    @pytest.mark.parametrize('attrs', [
        {'user_name': 'someuser', 'real_name': 'asdf hjkl'},
        {'callink_oid': 46187, 'is_group': True},
    ])
    def test_invalid_request_warning(
        self,
        fake_new_account_request,
        fake_credentials,
        mock_valid_calnet_uid,
        attrs,
    ):
        errors, warnings = validate_request(
            fake_new_account_request._replace(**attrs),
            fake_credentials,
            mock_session([[False]]),
        )
        assert warnings

    @pytest.mark.parametrize('attrs', [
        {'user_name': 'ckuehl'},
    ])
    def test_invalid_request_error(
        self,
        fake_new_account_request,
        fake_credentials,
        mock_valid_calnet_uid,
        attrs,
    ):
        errors, warnings = validate_request(
            fake_new_account_request._replace(**attrs),
            fake_credentials,
            mock_session([[False]]),
        )
        assert errors

    def test_invalid_request_already_submitted_submitted(
        self,
        fake_new_account_request,
        fake_credentials,
        mock_valid_calnet_uid,
    ):
        with mock.patch('ocflib.account.submission.username_pending', return_value=True):
            errors, warnings = validate_request(
                fake_new_account_request,
                fake_credentials,
                mock_session([[False]]),
            )
        assert errors

        with mock.patch('ocflib.account.submission.user_has_request_pending', return_value=True):
            errors, warnings = validate_request(
                fake_new_account_request,
                fake_credentials,
                mock_session([[False]]),
            )
        assert errors


class TestCreateAccount:

    @pytest.mark.parametrize('is_group,calnet_uid,callink_oid,expected', [
        (False, 123456, None, {'calnetUid': ['123456']}),
        (True, None, 123456, {'callinkOid': ['123456']}),
    ])
    def test_create(
        self,
        is_group,
        calnet_uid,
        callink_oid,
        expected,
        fake_new_account_request,
        fake_credentials
    ):
        @contextmanager
        def report_status(start, end, line):
            yield
        with mock.patch('ocflib.account.creation.create_kerberos_principal_with_keytab') as kerberos, \
                mock.patch('ocflib.account.creation.create_ldap_entry_with_keytab') as ldap, \
                mock.patch('ocflib.account.creation.create_home_dir') as home_dir, \
                mock.patch('ocflib.account.creation.create_web_dir') as web_dir, \
                mock.patch('ocflib.account.creation.send_created_mail') as send_created_mail, \
                mock.patch('ocflib.account.creation._get_first_available_uid', return_value=42):

            fake_new_account_request = fake_new_account_request._replace(
                is_group=is_group,
                calnet_uid=calnet_uid,
                callink_oid=callink_oid,
            )
            create_account(
                fake_new_account_request,
                fake_credentials,
                report_status,
            )
            kerberos.assert_called_once_with(
                fake_new_account_request.user_name,
                fake_credentials.kerberos_keytab,
                fake_credentials.kerberos_principal,
                password='hunter2000',
            )
            ldap.assert_called_once_with(
                'uid=someuser,ou=People,dc=OCF,dc=Berkeley,dc=EDU',
                dict({
                    'cn': ['Some User'],
                    'gidNumber': ['1000'],
                    'objectClass': ['ocfAccount', 'account', 'posixAccount'],
                    'uidNumber': ['42'],
                    'homeDirectory': ['/home/s/so/someuser'],
                    'loginShell': ['/bin/bash'],
                    'mail': ['some.user@ocf.berkeley.edu'],
                    'userPassword': ['{SASL}someuser@OCF.BERKELEY.EDU'],
                }, **expected),
                fake_credentials.kerberos_keytab,
                fake_credentials.kerberos_principal,
            )
            home_dir.assert_called_once_with(fake_new_account_request.user_name)
            web_dir.assert_called_once_with(fake_new_account_request.user_name)
            send_created_mail.assert_called_once_with(fake_new_account_request)
